from .UnitList import UnitStandard
from .QuantityClass import Quantity

class Unit:
    
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass
        
    def __new__(cls, unit=None):
        if unit:
            return Quantity(1,str(unit))
        else:
            return object.__new__(cls)
    
    def __getattr__(self, unit):
        return Quantity(1,str(unit))

class Constant:

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass
    
    def __new__(cls, unit=None):
        if unit:
            return Quantity(1,f"[{str(unit)}]")
        else:
            return object.__new__(cls)
    
    def __getattr__(self, unit):
        return Quantity(1,f"[{str(unit)}]")



import numpy as np
from dataclasses import dataclass, field, fields
from .FractionClass import Fraction

@dataclass
class BaseUnits:

    baseunits: dict = field(default_factory=dict)
    symbol: str = ':'

    def __post_init__(self):
        for unit,exp in self.baseunits.items():
            if not isinstance(exp, Fraction):
                self.baseunits[unit] = Fraction(exp)
        # remove zero dimensions
        for unit in list(self.baseunits.keys()):
            if self.baseunits[unit].num==0:
                del self.baseunits[unit]

    def __str__(self):
        baseunits = []
        for unit,exp in self.baseunits.items():
            if exp.num not in [0, -0]:
                unit = unit.replace(self.symbol,"")
                baseunits.append(f"{unit}={str(exp)}")
        baseunits = " ".join(baseunits)
        return f"BaseUnits({baseunits})"

    def __repr__(self):
        baseunits = []
        for unit,exp in self.baseunits.items():
            if exp.num not in [0, -0]:
                unit = unit.replace(self.symbol,"")
                baseunits.append(f"{unit}={str(exp)}")
        baseunits = " ".join(baseunits)
        return f"BaseUnits({baseunits})"

    def __add__(self, other):
        baseunits = dict(self.baseunits)
        for unit,exp in other.baseunits.items():
            baseunits[unit] = baseunits[unit]+exp if unit in baseunits else exp
        return BaseUnits(baseunits)
    
    def __sub__(self, other):
        baseunits = dict(self.baseunits)
        for unit,exp in other.baseunits.items():
            baseunits[unit] = baseunits[unit]-exp if unit in baseunits else -exp
        return BaseUnits(baseunits)

    def __mul__(self, other):
        baseunits = dict(self.baseunits)
        for unit,exp in self.baseunits.items():
            baseunits[unit] *= other
        return BaseUnits(baseunits)

    def __truediv__(self, div):
        baseunits = dict(self.baseunits)
        for unit,exp in self.baseunits.items():
            baseunits[unit] /= div
        return BaseUnits(baseunits)
    
    def expression(self):
        units = []
        for unit,exp in self.baseunits.items():
            symbol = unit.replace(self.symbol,'')
            if exp.num==1 and exp.den==1:
                units.append(symbol)
            else:
                units.append(f"{symbol}{exp}")
        return "*".join(units)
    
    def value(self):
        baseunits = {}
        for unit,exp in self.baseunits.items():
            baseunits[unit] = exp.value()
        return baseunits



from ..structs import ParameterDict
from enum import Enum
from dataclasses import dataclass

@dataclass
class ConstCGS:
    c: float     = 2.99792458e10     # speed of light            (cm s-1)            
    h: float     = 6.6260755e-27     # Planck constant           (erg s)             
    hbar: float  = 1.05457266e-27    # reduced Planck constant   (erg s)             
    G: float     = 6.67259e-8        # gravitational constant    (cm3 g-1 s-2)       
    e: float     = 4.8032068e-10     # electron charge           (esu)               
    m_e: float   = 9.1093897e-28     # electron mass             (g)                 
    m_p: float   = 1.6726231e-24     # proton mass               (g)                 
    m_n: float   = 1.6749286e-24     # neutron mass              (g)                 
    u: float     = 1.6605402e-24     # atomic mass unit          (g)                 
    k_B: float   = 1.380658e-16      # Boltzmann constant        (erg K-1)           
    eV: float    = 1.6021772e-12     # electron-Volt             (erg)               
    sigma: float = 5.67051e-5        # Stefan-Boltzmann constant (erg cm-2 K-4 s-1)   
    alpha: float = 7.29735308-3      # fine structure constant
    N_a: float   = 6.0221367e23      # Avogadro's number                             

@dataclass
class ConstMKS:
    c: float     = 299792.458        # speed of light            (km s-1)            
    h: float     = 6.62607015e-34    # Planck constant           (J s)             
    hbar: float  = 1.054571817e-34   # reduced Planck constant   (J s)             
    G: float     = 6.67430e10-11     # gravitational constant    (m3 kg-1 s-2)       
    e: float     = 1.602176634e-19   # electron charge           (C)
    m_e: float   = 9.1093837015e-31  # electron mass             (kg)
    m_p: float   = 1.67262192e-27    # proton mass               (kg)                 
    m_n: float   = 1.67492749804e-27 # neutron mass              (kg)                 
    u: float     = 1.66053906660e-27 # atomic mass unit          (kg)                 
    k_B: float   = 1.380649e-2       # Boltzmann constant        (J K-1)
    eV: float    = 1.602176634e-19   # electron-Volt             (J)
    sigma: float = 5.670374419e-8    # Stefan-Boltzmann constant (W m-2 K-4)   
    alpha: float = 7.29735308-3      # fine structure constant              
    N_a: float   = 6.0221367e23      # Avogadro's number                             



import numpy as np
import pytest
from math import isclose
import os
import sys
sys.path.insert(0, 'src')

from scinumtools.structs import ParameterDict
from scinumtools.phys.units.UnitList import UnitStandard, UnitPrefixes
from scinumtools.phys.units import *

def test_quantity():
    
    assert str(Quantity(123e2))         == "Quantity(1.230e+04)"
    q = Quantity(123e2, 'km/s')
    assert str(q) == "Quantity(1.230e+04 km*s-1)"
    assert str(q.value())               == "12300.0"
    assert str(q.value('m/s'))          == "12300000.0"
    assert str(q.value({'m':1,'s':-1})) == "12300000.0"
    assert str(q.units())               == "km*s-1"
    
    result = "Quantity(1.230e+04 m*s2:3)"
    assert str(Quantity(123e2, [1,0,(2,3),0,0,0,0,0] ))    == result
    assert str(Quantity(123e2, Dimensions(m=1, s=(2,3)) )) == result

    result = "Quantity(1.230e+04 J2*kg2:3)"
    assert str(Quantity(123e2, {'J': 2, 'kg':(2,3)} )) == result
    assert str(Quantity(123e2, BaseUnits({'J': 2, 'kg':(2,3)}) )) == result

    result = "Quantity(2.460e+04 m)"
    assert str(Quantity(123e2, Quantity(2, 'm'))) == result
    assert str(Quantity(123e2, 2*Unit('m'))) == result
    
def test_units():

    assert str(Unit('m'))                   == "Quantity(1.000e+00 m)"
    assert str(Unit('kg*m2*s-2'))           == "Quantity(1.000e+00 kg*m2*s-2)"
                                            
    unit = Unit()                           
    assert str(unit.m)                      == "Quantity(1.000e+00 m)"
    assert str(2*unit.kJ)                   == "Quantity(2.000e+00 kJ)"
    assert str(unit.kg*unit.m**2/unit.s**2) == "Quantity(1.000e+00 kg*m2*s-2)"
    assert str(unit.J.to('erg'))            == "Quantity(1.000e+07 erg)"
    assert str(unit.m**(2,3))               == "Quantity(1.000e+00 m2:3)"
    assert str(unit.m**Fraction(2,3))       == "Quantity(1.000e+00 m2:3)"
    assert str((9*unit.m)**(1,2))           == "Quantity(3.000e+00 m1:2)"

    with Unit() as unit:
        assert str(unit.m)                  == "Quantity(1.000e+00 m)"

def test_constants():

    assert str(Constant('c')) == "Quantity(1.000e+00 [c])"

    const = Constant()
    assert str(const.c)       == "Quantity(1.000e+00 [c])"
    assert str(const.m_e)     == "Quantity(1.000e+00 [m_e])"

    with Constant() as const:
        assert str(const.c)   == "Quantity(1.000e+00 [c])"
    
def test_aliases():
    from scinumtools import quant, unit, const
    assert str(quant(23, 'km'))   == "Quantity(2.300e+01 km)"
    assert str(unit('m'))         == "Quantity(1.000e+00 m)"
    assert str(const('c'))        == "Quantity(1.000e+00 [c])"    
    
def test_dimensions():

    # Test simplification
    dims = Dimensions(m=(5,-2), g=3, s=1, cd=(-0,3), K=(34,1), rad=(18,12))
    assert str(dims) == "Dimensions(m=-5:2 g=3 s=1 K=34 rad=3:2)"

    # Test arithmetics
    dims1 = Dimensions(m=3, g=(3,2))
    dims2 = Dimensions(m=2, g=(4,7))
    assert not dims1==dims2
    assert str(dims1+dims2) == "Dimensions(m=5 g=29:14)"
    assert str(dims1-dims2) == "Dimensions(m=1 g=13:14)"
    assert str(dims1*2)     == "Dimensions(m=6 g=3)"
    assert str(dims2*0.5)   == "Dimensions(m=1 g=2:7)"

    # Test values
    value = [3, (3,2), 0, 0, 0, 0, 0, 0]
    dims = Dimensions(*value)
    assert str(dims) == "Dimensions(m=3 g=3:2)" 
    assert dims.value() == value
    assert dims.value(dtype=dict) == {'m': 3, 'g':(3,2)}

def test_fractions():

    # Test initialization
    assert str(Fraction(2))   == "2"
    assert str(Fraction(0,3)) == "0"
    assert str(Fraction(1,3)) == "1:3"

    # Test arithmetics
    assert str(Fraction(1,3)+3)             == "10:3"
    assert str(Fraction(1,3)+(3,2))         == "11:6"
    assert str(Fraction(1,3)+Fraction(3,2)) == "11:6"
    assert str(Fraction(1,3)-3)             == "-8:3"
    assert str(Fraction(1,3)-(3,2))         == "-7:6"
    assert str(Fraction(1,3)-Fraction(3,2)) == "-7:6"
    assert str(Fraction(1,3)*3)             == "1"
    assert str(Fraction(1,3)*(3,2))         == "1:2"
    assert str(Fraction(1,3)*Fraction(3,2)) == "1:2"
    assert str(Fraction(1,3)/3)             == "1:9"
    assert str(Fraction(1,3)/(3,2))         == "2:9"
    assert str(Fraction(1,3)/Fraction(3,2)) == "2:9"
    
def test_base_units():

    # Test simplification
    bu = BaseUnits({'g':(3,2), 'km': 3, '[m_p]': (3,1)})
    assert str(bu) == "BaseUnits(g=3:2 km=3 [m_p]=3)"

    # Test arithmetics
    bu1 = BaseUnits({'km':3,'g':(3,2)})
    bu2 = BaseUnits({'km':2,'g':(4,7)})
    assert not bu1 == bu2
    assert str(bu1+bu2)  == "BaseUnits(km=5 g=29:14)"
    assert str(bu1-bu2)  == "BaseUnits(km=1 g=13:14)"
    assert str(bu1*2)    == "BaseUnits(km=6 g=3)"
    assert str(bu2*0.5)  == "BaseUnits(km=1 g=2:7)"

    # Test values
    value = {"km": 2, "K": (3,2)}
    bu = BaseUnits(dict(value))
    assert str(bu) == "BaseUnits(km=2 K=3:2)"
    assert bu.value() == value
    
def test_definitions():
    
    unitlist = ParameterDict(['magnitude','dimensions','definition','name'], UnitStandard)
    for symbol, unit in unitlist.items():
        if not isinstance(unit['definition'],str):
            continue
        q = Quantity(1, unit['definition'])
        assert isclose(q.magnitude,  unit['magnitude'], rel_tol=1e-07)
        assert q.dimensions.value() == unit['dimensions']
        
    prefixes = ParameterDict(['magnitude','dimensions','definition','name'], UnitPrefixes)
    for symbol, unit in prefixes.items():
        if not isinstance(unit['definition'],str):
            continue
        q = Quantity(1, unit['definition'])
        assert isclose(q.magnitude,  unit['magnitude'], rel_tol=1e-07)
        assert q.dimensions.value() == unit['dimensions']
        
def test_scalar_arithmetics():

    # addition and substtraction
    assert str(Quantity(1.1, 'km')-Quantity(100, 'm')) == "Quantity(1.000e+00 km)"
    assert str(Quantity(1.0, 'km')+Quantity(100, 'm')) == "Quantity(1.100e+00 km)"

    # multiplication, division and power
    q = Quantity(123e2, [3,3,0,0,1,0,0,0])
    assert str(q) == "Quantity(1.230e+04 m3*g3*C)"    
    q /= Quantity(123, 'C')
    assert str(q) == "Quantity(1.000e+02 m3*g3)"
    q *= Quantity(2, 's2')
    assert str(q) == "Quantity(2.000e+02 m3*g3*s2)"
    q = Quantity(123, "kg3*cm-2*s")
    assert str(q) == "Quantity(1.230e+02 kg3*cm-2*s)"
    q = Quantity(123e34, "J")
    assert str(q) == "Quantity(1.230e+36 J)"
    q /= Quantity(123, 's')
    assert str(q) == "Quantity(1.000e+34 J*s-1)"
    q = q ** 2
    assert str(q) == "Quantity(1.000e+68 J2*s-2)"
    q = q.to('kg2*m4/s6')
    assert str(q) == "Quantity(1.000e+68 kg2*m4*s-6)"
    q = Quantity(123, "kg3*s/(cm2*m)")
    assert str(q) == "Quantity(1.230e+02 kg3*s*cm-2*m-1)"
    q = q.to("kg3*s/cm3")
    assert str(q) == "Quantity(1.230e+00 kg3*s*cm-3)"
    q = Quantity(134e-34)
    assert str(q) == "Quantity(1.340e-32)"

    # error if converting different units
    with pytest.raises(Exception) as excinfo:
        q = q.to("kg3*s/cm3")
    assert excinfo.value.args[0]=="Converting units with different dimensions:"
    assert excinfo.value.args[1].value()==[0, 0, 0, 0, 0, 0, 0, 0]
    assert excinfo.value.args[2].value()==[-3, 3, 1, 0, 0, 0, 0, 0]

    # test unit conversion on scalars
    result = "Quantity(3.000e+03 m)"
    assert str(Quantity(3, 'km').to('m'))                 == result
    assert str(Quantity(3, 'km').to([1,0,0,0,0,0,0,0]))   == result
    assert str(Quantity(3, 'km').to(Dimensions(m=1)))     == result
    assert str(Quantity(3, 'km').to({'m':1}))             == result
    assert str(Quantity(3, 'km').to(BaseUnits({'m':1})))  == result
    assert str(Quantity(3, 'km').to(Unit().m))            == result

    # reset base units if dimensions are all zero
    assert str(Quantity(3, 'kg*m2/s2')/Quantity(2, 'J')) == "Quantity(1.500e+00)"
    
def test_array_arithmetics():

    # Test basic arithmetics
    q = Quantity([2,3,4], 'm')
    assert str(q) == "Quantity([2. 3. 4.] m)"
    q += Quantity(2, 'm')
    assert str(q) == "Quantity([4. 5. 6.] m)"
    q -= Quantity(2, 'm')
    assert str(q) == "Quantity([2. 3. 4.] m)"
    q /= Quantity(2)
    assert str(q) == "Quantity([1.  1.5 2. ] m)"
    q *= Quantity(2)
    assert str(q) == "Quantity([2. 3. 4.] m)"
    q /= 2
    assert str(q) == "Quantity([1.  1.5 2. ] m)"
    q *= 2
    assert str(q) == "Quantity([2. 3. 4.] m)"
    q /= [1, 1.5, 2]
    assert str(q) == "Quantity([2. 2. 2.] m)"
    q *= [1, 1.5, 2]
    assert str(q) == "Quantity([2. 3. 4.] m)"
    q = q ** 10
    assert str(q) == "Quantity([1.024e+03 5.905e+04 1.049e+06] m10)"

    # Test unit conversion on arrays
    assert str(Quantity([1,2,3], 'm').to('km')) == "Quantity([0.001 0.002 0.003] km)"

    # Array slicing
    q = Quantity([1,2,3], 'm')
    assert str(q[:2]) == "Quantity([1. 2.] m)"
    
def test_numpy():
    
    # Test numpy functions
    assert str(np.sqrt(Quantity(4, 'm2')))             == "Quantity(2.000e+00 m)"
    assert str(np.sqrt(Quantity([4, 9, 16], 'm2')))    == "Quantity([2. 3. 4.] m)"
    assert str(np.sqrt(Quantity([4, 9, 16], 'm3')))    == "Quantity([2. 3. 4.] m3:2)"
    assert str(np.sqrt(Quantity([4, 9, 16], 'm-3')))   == "Quantity([2. 3. 4.] m-3:2)"
    assert str(np.sqrt(Quantity([4, 9, 16], 'm2:3')))  == "Quantity([2. 3. 4.] m1:3)"
    assert str(np.sqrt(Quantity([4, 9, 16], 'm2:3*g3:5*s5')))  == "Quantity([2. 3. 4.] m1:3*g3:10*s5:2)"
    assert str(np.cbrt(Quantity([8, 27, 64], 'm3')))   == "Quantity([2. 3. 4.] m)"
    assert str(np.power(Quantity([2, 3, 4], 'm'),3))   == "Quantity([ 8. 27. 64.] m3)"
    assert str(np.sin(Quantity([45, 60], "deg")))      == "Quantity([0.707 0.866])"
    assert str(np.cos(Quantity([45, 60], "deg")))      == "Quantity([0.707 0.5  ])"
    assert str(np.tan(Quantity([45, 60], "deg")))      == "Quantity([1.    1.732])"
    assert str(np.arcsin(Quantity([0.3, -0.7])))       == "Quantity([ 0.305 -0.775] rad)"
    assert str(np.arccos(Quantity([0.3, -0.7])))       == "Quantity([1.266 2.346] rad)"
    assert str(np.arctan(Quantity([0.3, -0.7])))       == "Quantity([ 0.291 -0.611] rad)"
    assert str(np.linspace(0,Quantity(20,'m'),3))      == "Quantity([ 0. 10. 20.] m)"
    assert str(np.linspace(Quantity(10,'m'),Quantity(0.03,'km'),3))  == "Quantity([10. 20. 30.] m)"
    assert str(np.absolute(Quantity(-3,'m')))          == "Quantity(3.000e+00 m)"
    assert str(np.linspace(0,Quantity(23,'km'),3))     == "Quantity([ 0.  11.5 23. ] km)"
    assert str(np.logspace(1,Quantity(3,'m'),3))       == "Quantity([  10.  100. 1000.] m)"
    assert str(np.absolute(Quantity(-3,'m')))          == "Quantity(3.000e+00 m)"
    assert str(np.abs(Quantity(-3,'m')))               == "Quantity(3.000e+00 m)"

def test_operation_sides():
    
    p = Quantity([2,3,4], 'm')
    q = Quantity([5,6,7], 'm')
    assert p+q == q+p
    assert p-q == -(q-p)
    assert q*2 == 2*q
    assert p/2 == 1/(2/p)
    
def test_unique_names():

    units = list(UnitStandard.keys())
    for unit in UnitStandard.keys():
        for prefix in UnitPrefixes.keys():
            units.append(f"{prefix}{unit}")
    units.sort()
    seen = set()
    dupes = [x for x in units if x in seen or seen.add(x)]    
    assert len(dupes)==0

def test_temperatures():
    
    # Temperature conversions
    assert str(Quantity(23,'K').to('Cel'))        == "Quantity(-2.501e+02 Cel)"
    assert str(Quantity(23,'K').to('degF'))       == "Quantity(-4.183e+02 degF)"
    assert str(Quantity(23,'K').to('degR'))       == "Quantity(4.140e+01 degR)"
    assert str(Quantity(23,'Cel').to('K'))        == "Quantity(2.961e+02 K)"
    assert str(Quantity(23,'Cel').to('degF'))     == "Quantity(7.340e+01 degF)"
    assert str(Quantity(23,'Cel').to('degR'))     == "Quantity(5.331e+02 degR)"
    assert str(Quantity(23,'degF').to('K'))       == "Quantity(2.681e+02 K)"
    assert str(Quantity(23,'degF').to('Cel'))     == "Quantity(-5.000e+00 Cel)"
    assert str(Quantity(23,'degF').to('degR'))    == "Quantity(4.827e+02 degR)"
    assert str(Quantity(23,'degR').to('K'))       == "Quantity(1.278e+01 K)"
    assert str(Quantity(23,'degR').to('Cel'))     == "Quantity(-2.604e+02 Cel)"
    assert str(Quantity(23,'degR').to('degF'))    == "Quantity(-4.367e+02 degF)"
    assert str(Quantity(2300,'Cel').to('kK'))     == "Quantity(2.573e+00 kK)"
    
def test_inversion():

    assert str(Quantity(23, 'Hz').to('s'))        == "Quantity(4.348e-02 s)"
    assert str(Quantity(34, 'Ohm').to('S'))       == "Quantity(2.941e-02 S)"
    assert str(Quantity(102, 'J').to('erg-1'))    == "Quantity(9.804e-10 erg-1)"



import numpy as np
import os
import sys
sys.path.insert(0, 'src')

from scinumtools.phys import *

def test_constants():

    assert ConstCGS.k_B == 1.380658e-16
    assert ConstMKS.k_B == 1.380649e-2 
    



import numpy as np
import pandas as pd
from textwrap import dedent
import sys
sys.path.insert(0, 'src')

from scinumtools.structs import *

def test_row_collector_list():
    
    columns = ['col1','col2','col3']
    with ListCollector(columns) as lc:
        lc.append([1,2,3])
        lc.append([4,5,6])
        lc.append([7,8,9])
        assert lc.col1 == [1,4,7]
        assert lc.col2 == [2,5,8]
        assert lc.col3 == [3,6,9]
        assert lc.to_dict() == dict(
            col1 = [1,4,7],
            col2 = [2,5,8],
            col3 = [3,6,9],
        )
        pd.testing.assert_frame_equal(
            lc.to_dataframe(),
            pd.DataFrame(dict(
                col1 = [1,4,7],
                col2 = [2,5,8],
                col3 = [3,6,9],
            ))
        )
        result = dedent("""\
               col1  col2  col3
            0     1     2     3
            1     4     5     6
            2     7     8     9
        """.rstrip())
        assert lc.to_text() == result
        assert str(lc) == result
        
def test_row_collector_array():
    
    with ArrayCollector(['col1','col2','col3']) as ac:
        ac.append([1,2,3])
        ac.append([4,5,6])
        ac.append([7,8,0])
        np.testing.assert_equal(ac.col1, [1,4,7])
        np.testing.assert_equal(ac.col2, [2,5,8])
        np.testing.assert_equal(ac.col3, [3,6,0])
        np.testing.assert_equal(ac.to_dict(), dict(
            col1 = [1,4,7], 
            col2 = [2,5,8],
            col3 = [3,6,0],
        ))
        pd.testing.assert_frame_equal(
            ac.to_dataframe(),
            pd.DataFrame(dict(
                col1 = [1.,4.,7.],
                col2 = [2.,5.,8.],
                col3 = [3.,6.,0.],
            ))
        )
        result = dedent("""\
               col1  col2  col3
            0   1.0   2.0   3.0
            1   4.0   5.0   6.0
            2   7.0   8.0   0.0
        """.rstrip())
        assert ac.to_text() == result
        assert str(ac) == result
        
    columns = {'col1':dict(dtype=str),'col2':dict(dtype=float),'col3':dict(dtype=bool)}
    with ArrayCollector(columns) as ac:
        ac.append([1,2,3])
        ac.append([4,5,6])
        ac.append([7,8,0])
        np.testing.assert_equal(ac.col1, ['1','4','7'])
        np.testing.assert_equal(ac.col2, [2,5,8])
        np.testing.assert_equal(ac.col3, [True,True,False])
        np.testing.assert_equal(ac.to_dict(), dict(
            col1 = ['1','4','7'], 
            col2 = [2,5,8],
            col3 = [True,True,False],
        ))
        pd.testing.assert_frame_equal(
            ac.to_dataframe(),
            pd.DataFrame(dict(
                col1 = ['1','4','7'],
                col2 = [2.,5.,8.],
                col3 = [True,True,False],
            ))
        )
        result = dedent("""\
              col1  col2   col3
            0    1   2.0   True
            1    4   5.0   True
            2    7   8.0  False
        """.rstrip())
        assert ac.to_text() == result
        assert str(ac) == result

def test_parameter_list():
    def test(params):
        assert params[0].a == 1
        assert params[1]['a'] == 4
        for param in params:
            assert param['a'] in [1,4]
        for index,settings in params.items():
            assert param['b'] in [2,5]        
    # append values
    with ParameterList(['a','b','c']) as params:
        params.append([1, 2, 3])
        params.append([4, 5, 6])
        test(params)
    # direct value insertion
    params = ParameterList(['a','b','c'],[
        [1, 2, 3],
        [4, 5, 6]
    ])
    test(params)

def test_parameter_dict():
    def test(params):
        assert params['d'].a == 1
        assert params['e']['a'] == 4
        for param in params:
            assert param['a'] in [1,4]
        for index,settings in params.items():
            assert param['b'] in [2,5]
    # append values
    with ParameterDict(['a','b','c']) as params:
        params['d'] = [1, 2, 3]
        params.append( 'e', [4, 5, 6] )
        test(params)
    # direct value insertion
    params = ParameterDict(['a','b','c'],{
        'd': [1, 2, 3],
        'e': [4, 5, 6]
    })
    test(params)



import numpy as np
import os
import sys
sys.path.insert(0, 'src')

from scinumtools.data import *

def test_caching():

    file_cache = "tests/cached_data.npy"
    
    if os.path.isfile(file_cache):
        os.remove(file_cache)
    assert not os.path.isfile(file_cache)
    
    @CachedFunction(file_cache)
    def read_data(a, b):
        return dict(a=a, b=b)

    data = read_data('foo','bar')    
    assert data == dict(a='foo', b='bar')

    data = read_data('foo2','bar2')
    assert os.path.isfile(file_cache)
    assert data == dict(a='foo', b='bar')
    
    if os.path.isfile(file_cache):
        os.remove(file_cache)
    assert not os.path.isfile(file_cache)

    @CachedFunction(file_cache)
    def read_data():
        return {
            'array': np.linspace(1,100,100),
            'scalar': 34,
            'string': 'foo',
        }
    
    data = read_data()
    np.testing.assert_equal(data, dict(
        array=np.linspace(1,100,100),
        scalar=34,
        string='foo',
    ))
    
def test_normalizing():

    xlen = 10
    ylen = 20
    data = np.linspace(1,xlen*ylen,xlen*ylen).reshape(xlen,ylen) - 10

    # Test only data
    with NormalizeData() as n:
        for row in data:
            n.append(row)
        linnorm = n.linnorm()
        lognorm = n.lognorm()
        zranges = n.zranges()

    assert linnorm.vmin == -9.
    assert linnorm.vmax == 190.0
    assert lognorm.vmin == 0
    assert lognorm.vmax == 2.278753600952829
    assert zranges.minpos == 1
    assert zranges.min == -9
    assert zranges.max == 190.0
    
    # Test data and x-axis
    with NormalizeData(xaxis=True) as n:
        for r,row in enumerate(data):
            xdata = np.linspace(-r,r,xlen)
            n.append(row, xdata)
        xranges = n.xranges()

    assert xranges.minpos == 0.11111111111111116
    assert xranges.min == -9
    assert xranges.max == 9

    # Test data and both axes
    with NormalizeData(xaxis=True, yaxis=True) as n:
        for r,row in enumerate(data):
            xdata = np.linspace(-r,r,xlen)
            ydata = np.linspace(-r*2,r*2,ylen)
            n.append(row, xdata, ydata)
        xranges = n.xranges()
        yranges = n.yranges()

    assert xranges.minpos == 0.11111111111111116
    assert xranges.min == -9
    assert xranges.max == 9
    assert yranges.minpos == 0.10526315789473673
    assert yranges.min == -18
    assert yranges.max == 18
    
def test_list_to_grid():

    data = range(5)
    ncols = 2
    
    # List all defined grid points
    grid = [
        (0, 0, 0, 0),
        (1, 0, 1, 1),
        (2, 1, 0, 2),
        (3, 1, 1, 3),
        (4, 2, 0, 4),
    ]
    for r,row in enumerate(ListToGrid(data,ncols)):
        assert grid[r] == row

    # List all missing grid points
    for r,row in enumerate(ListToGrid(data,ncols,missing=True)):
        assert (5, 2, 1) == row

    # List transposed grid points
    grid = [
        (0, 0, 0, 0),
        (1, 1, 0, 1),
        (2, 2, 0, 2),
        (3, 0, 1, 3),
        (4, 1, 1, 4),
    ]
    for r,row in enumerate(ListToGrid(data,ncols,transpose=True)):
        assert grid[r] == row
        
    # List all missing transposed grid points
    for r,row in enumerate(ListToGrid(data,ncols,transpose=True,missing=True)):
        assert (5, 2, 1) == row
        
def test_dict_to_grid():

    data = dict(
        a = 0,
        b = 1,
        c = 2,
        d = 3,
        e = 4
    )
    ncols = 2
    
    # List all defined grid points
    grid = [
        (0, 0, 0, 'a', 0),
        (1, 0, 1, 'b', 1),
        (2, 1, 0, 'c', 2),
        (3, 1, 1, 'd', 3),
        (4, 2, 0, 'e', 4),
    ]
    for r,row in enumerate(DictToGrid(data,ncols)):
        assert grid[r] == row

    # List all missing grid points
    grid = [
        (5, 2, 1)
    ]
    for r,row in enumerate(DictToGrid(data,ncols,missing=True)):
        assert grid[r] == row
        
    # List all transposed grid points
    grid = [
        (0, 0, 0, 'a', 0),
        (1, 1, 0, 'b', 1),
        (2, 2, 0, 'c', 2),
        (3, 0, 1, 'd', 3),
        (4, 1, 1, 'e', 4),
    ]
    for r,row in enumerate(DictToGrid(data,ncols,transpose=True)):
        assert grid[r] == row

    # List all missing transposed grid points
    grid = [
        (5, 2, 1)
    ]
    for r,row in enumerate(DictToGrid(data,ncols,transpose=True,missing=True)):
        assert grid[r] == row
        
def test_thumbnail():

    # Test grayscale case
    nx, ny = 300, 200
    data = np.zeros((nx,ny))
    for i in range(nx):
        for j in range(ny):
            data[i,j] = (i-nx/2)**2 + (j-ny/2)**2
            imold = ThumbnailImage(
                extent = (-30,30,-20,20),
                data = data
            )
    np.testing.assert_equal(np.asarray(imold.im), data)

    # Test grayscale resizing
    data_resized = [
        [9.6480850e+03, 1.1210736e+04, 9.5458672e+03, 1.1227712e+04, 9.3136709e+03],
        [1.0238771e+04, 9.7449180e+03, 6.8944453e+03, 9.7642842e+03, 9.8572793e+03],
        [5.1896011e+03, 2.6150837e+03, 1.7226636e-03, 2.6263579e+03, 4.9675088e+03],
        [1.0238771e+04, 9.7449180e+03, 6.8944453e+03, 9.7642842e+03, 9.8572793e+03],
        [9.3407617e+03, 1.0838822e+04, 9.2195254e+03, 1.0855267e+04, 9.0168174e+03],
    ]    
    imnew = imold.resize((-35,35,-25,25), (5,5))
    np.testing.assert_almost_equal(np.asarray(imnew.im), data_resized, decimal=3)

    # Test RGB case
    nx, ny = 300, 200
    data = np.zeros((nx,ny,3), dtype=float)
    for i in range(nx):
        for j in range(ny):
            data[i,j] = [(i-nx/2)**2 + (j-ny/2)**2]*3
    data = (255*data/data.max()).astype(np.uint8)
    imold = ThumbnailImage(
        extent = (-30,30,-20,20),
        data = data,
        mode = 'RGB'
    )
    np.testing.assert_equal(np.asarray(imold.im), data)

    # Test RGB resizing
    data_resized = [
        [[76, 76, 76],
         [88, 88, 88],
         [75, 75, 75],
         [88, 88, 88],
         [73, 73, 73]],
        [[80, 80, 80],
         [76, 76, 76],
         [54, 54, 54],
         [76, 76, 76],
         [77, 77, 77]],
        [[40, 40, 40],
         [20, 20, 20],
         [ 0,  0,  0],
         [20, 20, 20],
         [39, 39, 39]],
        [[80, 80, 80],
         [76, 76, 76],
         [54, 54, 54],
         [76, 76, 76],
         [77, 77, 77]],
        [[73, 73, 73],
         [85, 85, 85],
         [72, 72, 72],
         [85, 85, 85],
         [71, 71, 71]],
    ]
    imnew = imold.resize((-35,35,-25,25), (5,5))
    np.testing.assert_equal(np.asarray(imnew.im), data_resized)
    



from matplotlib.colors import Normalize, LogNorm
from dataclasses import dataclass
import numpy as np

from ..structs import ArrayCollector

@dataclass
class Ranges:
    """ Dataclass that contains data ranges
    """
    minpos: float
    min: float
    max: float

class NormalizeData:
    """ Normalize numerical data across multiple plots
    """

    xaxis: bool
    yaxis: bool
    _collector: ArrayCollector
    
    def __enter__(self):
        return self
    
    def __exit__(self, type, value, tb):
        pass

    def __init__(self, xaxis=None, yaxis=None):
        self.xaxis = xaxis
        self.yaxis = yaxis
        columns = ['zminpos','zmin','zmax']
        if xaxis: columns += ['xminpos','xmin','xmax']
        if yaxis: columns += ['yminpos','ymin','ymax']
        self._collector = ArrayCollector(columns)

    def append(self, vdata, xdata=None, ydata=None):
        """ Append data to the range collector
        
        Statistics of x and y axes is optional

        :param vdata: Numerical data
        :param xdata: X-axis points
        :param ydata: Y-axis points
        """
        zminpos = np.nanmin(vdata[vdata>0]) if np.sum(vdata>0) else np.nan
        ranges = [zminpos,np.min(vdata),np.max(vdata)]
        if self.xaxis:
            if xdata is None:
                raise Exception("Missing x-axes values")
            xminpos = np.nanmin(xdata[xdata>0]) if np.sum(xdata>0) else np.nan
            ranges += [xminpos,np.min(xdata),np.max(xdata)]
        if self.yaxis:
            if ydata is None:
                raise Exception("Missing y-axes values")
            yminpos = np.nanmin(ydata[ydata>0]) if np.sum(ydata>0) else np.nan
            ranges += [yminpos,np.min(ydata),np.max(ydata)]
        self._collector.append(ranges)

    def data(self):
        """ Return collected ranges as a dictionary
        """
        return self._collector.to_dict()

    def xranges(self):
        """ Return x-axis ranges
        """
        return Ranges(
            minpos = np.nanmin(self._collector.xminpos),
            min = np.nanmin(self._collector.xmin),
            max = np.nanmax(self._collector.xmax)
        )
    
    def yranges(self):
        """ Return y-axis ranges
        """
        return Ranges(
            minpos = np.nanmin(self._collector.yminpos),
            min = np.nanmin(self._collector.ymin),
            max = np.nanmax(self._collector.ymax)
        )
        
    def zranges(self):
        """ Return data ranges
        """
        return Ranges(
            minpos = np.nanmin(self._collector.zminpos),
            min = np.nanmin(self._collector.zmin),
            max = np.nanmax(self._collector.zmax)
        )
    
    def linnorm(self):
        """ Return linear norm from ranges
        """
        return Normalize(
            vmin=np.nanmin(self._collector.zmin), 
            vmax=np.nanmax(self._collector.zmax)
        )

    def lognorm(self):
        """ Return logarithmic norm from ranges
        """
        return LogNorm(
            vmin=np.log10(np.nanmin(self._collector.zminpos)), 
            vmax=np.log10(np.nanmax(self._collector.zmax))
        )

def ListToGrid(data, ncols, missing=None, transpose=False):
    """
    Enumerate given data with an index and row/column number specified by number of columns

    :param list data: Any iterable data, e.g. list, array
    :param int ncols: Number of grid columns
    :param bool missing: List only missing grid points
    :param bool transpose: Transpose grid layout from horizontal to vertical arrangement
    """
    ndata = len(data)
    nrows = int(np.ceil(ndata/ncols))
    if missing:
        for i in range(ndata, ncols*nrows):
            yield (i,int(i%nrows),int(i/nrows)) if transpose else (i,int(i/ncols),int(i%ncols))
    else:
        for i,d in enumerate(data):
            yield (i,int(i%nrows),int(i/nrows),d) if transpose else (i,int(i/ncols),int(i%ncols),d)
    
def DictToGrid(data, ncols, missing=None, transpose=False):
    """
    Enumerate given data with an index and row/column number specified by number of columns

    :param dict data: A Python dictionary with keys and values
    :param int ncols: Number of grid columns
    :param bool missing: List only missing grid points
    :param bool transpose: Transpose grid layout from horizontal to vertical arrangement
    """
    ndata = len(data)
    nrows = int(np.ceil(ndata/ncols))
    if missing:
        for i in range(ndata, ncols*nrows):
            yield (i,int(i%nrows),int(i/nrows)) if transpose else (i,int(i/ncols),int(i%ncols))
    else:
        for i,(k,v) in enumerate(data.items()):
            yield (i,int(i%nrows),int(i/nrows),k,v) if transpose else (i,int(i/ncols),int(i%ncols),k,v)
    



import numpy as np
from dataclasses import dataclass, field, fields
from typing import Union
from .FractionClass import Fraction
    
@dataclass
class Dimensions:
    
    m: Union[int,Fraction] = field(default=0)
    g: Union[int,Fraction] = field(default=0)
    s: Union[int,Fraction] = field(default=0)
    K: Union[int,Fraction] = field(default=0)
    C: Union[int,Fraction] = field(default=0)
    cd: Union[int,Fraction] = field(default=0)
    mol: Union[int,Fraction] = field(default=0)
    rad: Union[int,Fraction] = field(default=0)
    
    def __post_init__(self):
        for f in fields(self):
            value = getattr(self, f.name)
            if not isinstance(value, Fraction):
                setattr(self, f.name, Fraction(value))

    def __str__(self):
        dimensions = []
        for f in fields(self):
            value = getattr(self, f.name)
            if value.num not in [0, -0]:
                dimensions.append(f"{f.name}={str(value)}")
        dimensions = " ".join(dimensions)
        return f"Dimensions({dimensions})"

    def __repr__(self):
        dimensions = []
        for f in fields(self):
            value = getattr(self, f.name)
            if value.num not in [0, -0]:
                dimensions.append(f"{f.name}={str(value)}")
        dimensions = " ".join(dimensions)
        return f"Dimensions({dimensions})"
                            
    def __add__(self, other):
        dimensions = {}
        for f in fields(self):
            if isinstance(other, Dimensions):
                dimensions[f.name] = getattr(self, f.name) + getattr(other, f.name)
            else:
                dimensions[f.name] = getattr(self, f.name) + other
        return Dimensions(**dimensions)
            
    def __sub__(self, other):
        dimensions = {}
        for f in fields(self):
            if isinstance(other, Dimensions):
                dimensions[f.name] = getattr(self, f.name) - getattr(other, f.name)
            else:
                dimensions[f.name] = getattr(self, f.name) - other
        return Dimensions(**dimensions)

    def __mul__(self, other):
        dimensions = {}
        for f in fields(self):
            dimensions[f.name] = getattr(self, f.name) * other
        return Dimensions(**dimensions)

    def __truediv__(self, other):
        dimensions = {}
        for f in fields(self):
            dimensions[f.name] = getattr(self, f.name) / other
        return Dimensions(**dimensions)
    
    def __eq__(self, other):
        for f in fields(self):
            if not getattr(self, f.name) == getattr(other, f.name):
                return False
        return True

    def __neg__(self):
        """ Inverse dimensions
        """
        dimensions = {}
        for f in fields(self):
            dimensions[f.name] = getattr(self, f.name) * -1
        return Dimensions(**dimensions)

    def value(self, dtype=list):
        if dtype==list:
            dimensions = []
            for f in fields(self):
                dimensions.append( getattr(self, f.name).value() )
        elif dtype==dict:
            dimensions = {}
            for f in fields(self):
                fraction = getattr(self, f.name)
                if fraction.num not in [0, -0]:
                    dimensions[f.name] = fraction.value()
        return dimensions



from dataclasses import dataclass
from typing import Union

@dataclass
class ParameterSettings:
    """ ParameterSettings class contain all settings of a parameter
    """
    
    def __enter__(self):
        return self
    
    def __exit__(self, type, value, tb):
        pass

    def __getitem__(self, key):
        return getattr(self,key)
    
    def __init__(self, settings):
        for key,value in settings.items():
            setattr(self,key,value)

@dataclass
class ParameterList:
    """ ParameterList class collects parameters in a concise form and create a list
    """
    
    _settings: list
    _data: list

    def __init__(self, settings: list, parameters: list=None):
        self._settings = settings
        self._data = []
        if parameters:
            for values in parameters:
                self.append(values)            

    def __enter__(self):
        return self
    
    def __exit__(self, type, value, tb):
        pass
    
    def __getitem__(self, key):
        return self._data[key]
    
    def append(self, values: list): 
        if self._data is None:
            self._data = []
        settings = ParameterSettings(dict(zip(self._settings, values)))
        self._data.append( settings )

    def items(self):
        return [(key,value) for key,value in enumerate(self._data)]

@dataclass
class ParameterDict:
    """ ParameterDict class collects parameters in a concise form and create a dictionary
    """
    _keys: list
    _settings: list
    _data: dict

    def __init__(self, settings: list, parameters: dict=None):
        self._settings = settings
        self._data = {}
        self._keys = []
        if parameters:
            for key,values in parameters.items():
                self.append(key, values)            

    def __enter__(self):
        return self
    
    def __exit__(self, type, value, tb):
        pass
    
    def __getitem__(self, key):
        if isinstance(key,int):
            return self._data[self._keys[key]]
        else:
            return self._data[key]

    def __setitem__(self, key, values):
        self.append(key, values)

    def keys(self):
        return self._keys
        
    def items(self):
        return self._data.items()

    def append(self, key, values):
        if self._data is None:
            self._data = {}
        settings = ParameterSettings(dict(zip(self._settings, values)))
        self._keys.append(key)
        self._data[key] = settings
        



class TemperatureConverter:
    convertable: bool = False
    unit1: dict = None
    unit2: dict = None
    
    def __enter__(self):
        return self
    
    def __exit__(self, type, value, traceback):
        pass
        
    def __init__(self, baseunits1, baseunits2):
        convertable = ['Cel','[degF]']
        for unitid, exponent in baseunits1.items():
            if unitid.split(':')[-1] in convertable:
                self.convertable = True
        for unitid, exponent in baseunits2.items():
            if unitid.split(':')[-1] in convertable:
                self.convertable = True
        if self.convertable:
            if len(baseunits1)!=1 or len(baseunits2)!=1:
                raise Exception("Only direct temperatures can be converted between each other:", baseunits1, baseunits2)
            else:
                self.unit1 = list(baseunits1.keys())[0].split(':')[-1]
                self.unit2 = list(baseunits2.keys())[0].split(':')[-1]
                
    def convert(self, magnitude1, magnitude2):
        if self.unit1=='K' and self.unit2=='Cel':
            return (magnitude1-273.15)/magnitude2
        elif self.unit1=='K' and self.unit2=='[degF]':
            return ((magnitude1-273.15)*9/5+32)/magnitude2
        elif self.unit1=='[degR]' and self.unit2=='[degF]':
            return (magnitude1*9/5-459.67)/magnitude2
        elif self.unit1=='[degR]' and self.unit2=='Cel':
            return ((magnitude1*9/5-491.67)*5/9)/magnitude2
        elif self.unit1=='Cel' and self.unit2=='K':
            return (magnitude1+273.15)/magnitude2
        elif self.unit1=='Cel' and self.unit2=='[degF]':
            return ((magnitude1*9/5)+32)/magnitude2
        elif self.unit1=='Cel' and self.unit2=='[degR]':
            return ((magnitude1*9/5)+491.67)/(magnitude2*9/5)
        elif self.unit1=='[degF]' and self.unit2=='K':
            return (((magnitude1-32)*5/9)+273.15)/magnitude2
        elif self.unit1=='[degF]' and self.unit2=='Cel':
            return ((magnitude1-32)*5/9)/magnitude2
        elif self.unit1=='[degF]' and self.unit2=='[degR]':
            return (magnitude1+459.67)/(magnitude2*9/5)



from typing import List
import numpy as np
import re
from typing import Union

from ...structs.ParameterClass import ParameterDict
from ...math.solver import ExpressionSolver, AtomBase, OperatorPar, OperatorMul, OperatorTruediv
from .UnitList import *
from .UnitConverters import *
from .DimensionsClass import Dimensions
from .BaseUnitsClass import BaseUnits
from .FractionClass import Fraction

HANDLED_FUNCTIONS = {}

class Quantity:
    prefixes: dict            # list of prefixes 
    unitlist: dict            # list of units
    
    magnitude: float          # quantity magnitude
    dimensions: Dimensions    # quantity dimensions
    baseunits: BaseUnits      # base units

    precision: float = 1e-7

    def __init__(
            self, magnitude:float,
            dimensions: Union[str,list,np.ndarray,Dimensions,dict,BaseUnits] = None,
            baseunits: Union[dict,BaseUnits] = None
    ):
        # Initialize settings
        self.unitlist = ParameterDict(['magnitude','dimensions','definition','name'], UnitStandard)
        self.prefixes = ParameterDict(['magnitude','dimensions','definition','name'], UnitPrefixes)
        if isinstance(magnitude, (int,float)):
            self.magnitude = float(magnitude)
        elif isinstance(magnitude, list):
            self.magnitude = np.array(magnitude, dtype=float)
        elif isinstance(magnitude, np.ndarray):
            self.magnitude = magnitude.astype(float)
        else:
            raise Exception("Magnitude can be either a number or an list/array of numbers")
        # Set quantity
        if dimensions is None:
            self.dimensions = Dimensions()
            self.baseunits = BaseUnits()
        elif isinstance(dimensions, (str, dict, BaseUnits)):
            if isinstance(dimensions, dict):
                dimensions = BaseUnits(dimensions).expression()
            elif isinstance(dimensions, BaseUnits):
                dimensions = dimensions.expression()
            with ExpressionSolver(self._atom_parser, [OperatorPar,OperatorMul,OperatorTruediv]) as es:
                unit = es.solve(dimensions)
            self.magnitude *= unit.magnitude
            self.dimensions = unit.dimensions
            self.baseunits = unit.baseunits
        elif isinstance(dimensions, Quantity):
            self.magnitude *= dimensions.magnitude
            self.dimensions = dimensions.dimensions
            self.baseunits = dimensions.baseunits
        elif isinstance(dimensions, (list, np.ndarray, Dimensions)):
            if isinstance(dimensions, Dimensions):
                self.dimensions = dimensions
            else:
                self.dimensions = Dimensions(*dimensions)
            if isinstance(baseunits, dict):
                self.baseunits = BaseUnits(baseunits)
            elif isinstance(baseunits, BaseUnits):
                self.baseunits = baseunits
            else:
                self.baseunits = BaseUnits(self.dimensions.value(dtype=dict))
        else:
            raise Exception("Insufficient quantity definition", magnitude, dimensions, baseunits)
        if self.dimensions == Dimensions():
            self.baseunits = BaseUnits()

    def _add(self, left, right):
        if not isinstance(left, Quantity):
            left = Quantity(left)
        if not isinstance(right, Quantity):
            right = Quantity(right)
        if not left.dimensions == right.dimensions:
            raise Exception('Dimension does not match:', left.dimensions, right.dimensions)
        magnitude = left.magnitude + right.to(left.baseunits).magnitude 
        dimensions = left.dimensions
        baseunits = left.baseunits
        return Quantity(magnitude, dimensions, baseunits)

    def __add__(self, other):
        return self._add(self, other)
    
    def __radd__(self, other):
        return self._add(other, self)

    def _sub(self, left, right):
        if not isinstance(left, Quantity):
            left = Quantity(left)
        if not isinstance(right, Quantity):
            right = Quantity(right)
        if not left.dimensions == right.dimensions:
            raise Exception('Dimension does not match:', left.dimensions, right.dimensions)
        magnitude = left.magnitude - right.to(left.baseunits).magnitude
        dimensions = left.dimensions
        baseunits = left.baseunits
        return Quantity(magnitude, dimensions, baseunits)

    def __sub__(self, other):
        return self._sub(self, other)

    def __rsub__(self, other):
        return self._sub(other, self)
    
    def _mul(self, left, right):
        if not isinstance(left, Quantity):
            left = Quantity(left)
        if not isinstance(right, Quantity):
            right = Quantity(right)
        magnitude = left.magnitude * right.magnitude
        dimensions = left.dimensions + right.dimensions
        baseunits = left.baseunits + right.baseunits
        return Quantity(magnitude, dimensions, baseunits)

    def __mul__(self, other):
        return self._mul(self, other)

    def __rmul__(self, other):
        return self._mul(other, self)
    
    def _truediv(self, left, right):
        if not isinstance(left, Quantity):
            left = Quantity(left)
        if not isinstance(right, Quantity):
            right = Quantity(right)
        magnitude = left.magnitude / right.magnitude
        dimensions = left.dimensions - right.dimensions
        baseunits = left.baseunits - right.baseunits
        return Quantity(magnitude, dimensions, baseunits)

    def __truediv__(self, other):
        return self._truediv(self, other)
    
    def __rtruediv__(self, other):
        return self._truediv(other, self)
    
    def __pow__(self, power: Union[float,int,tuple,Fraction]):
        if isinstance(power, tuple):
            exp = power[0]/power[1]
        elif isinstance(power, Fraction):
            exp = power.num/power.den
        else:
            exp = power
        magnitude = self.magnitude**exp
        dimensions = self.dimensions*power
        baseunits = self.baseunits*power
        return Quantity(magnitude, dimensions, baseunits)

    def __neg__(self):
        return Quantity(-self.magnitude, self.dimensions, self.baseunits)
    
    def __eq__(self, other):
        if not np.allclose(self.magnitude, other.magnitude, rtol=self.precision):
            return False
        if not self.dimensions==other.dimensions:
            return False
        return True
    
    def __str__(self):
        magnitude = self.value()
        if isinstance(magnitude, np.ndarray):
            with np.printoptions(precision=3, suppress=False, threshold=5):
                magnitude = f"{str(magnitude):s}"
        else:
            magnitude = f"{magnitude:.03e}"
        baseunits = self.baseunits.expression()
        if baseunits:
            return f"Quantity({magnitude:s} {baseunits})"
        else:
            return f"Quantity({magnitude:s})"
            
    def __repr__(self):
        magnitude = self.value()
        if isinstance(magnitude, np.ndarray):
            with np.printoptions(precision=3, suppress=False, threshold=5):
                magnitude = f"{str(magnitude):s}"
        else:
            magnitude = f"{magnitude:.03e}"
        baseunits = self.baseunits.expression()
        if baseunits:
            return f"Quantity({magnitude:s} {baseunits})"
        else:
            return f"Quantity({magnitude:s})"

    def __getitem__(self, key):
        return Quantity(self.magnitude[key], self.dimensions, self.baseunits)
        
    def __array__(self):
        return np.array(self.magnitude)
    
    def __array_prepare__(self, array, context=None):
        if context:
            if context[0] in [np.sin, np.cos, np.tan]:
                return np.array(context[1][0].to('rad'))
        return array
    
    def __array_wrap__(self, out_arr, context=None):
        if out_arr.ndim==0:
            out_arr = float(out_arr)
        dimensions = self.dimensions
        baseunits = self.baseunits
        if context:
            fn = context[0]
            if fn==np.sqrt:
                return Quantity(out_arr, dimensions/2, baseunits/2)
            elif fn==np.cbrt:
                return Quantity(out_arr, dimensions/3, baseunits/3)
            elif fn==np.power:
                return Quantity(out_arr, dimensions*context[1][1], baseunits*context[1][1])
            elif fn in [np.sin, np.cos, np.tan]:
                return Quantity(out_arr)
            elif fn in [np.arcsin, np.arccos, np.arctan]:
                return Quantity(out_arr, 'rad')
        return Quantity(out_arr, dimensions, baseunits)
    
    def __array_function__(self, func, types, args, kwargs):
        if func not in HANDLED_FUNCTIONS:
            raise NotImplementedError()
        return HANDLED_FUNCTIONS[func](*args, **kwargs)

    def _atom_parser(self, string=None):
        # parse number
        m = re.match(r'^[-]?([0-9.]+)(e([0-9+-]+)|)$', str(string))
        if m:
            magnitude = float(string)
            return Quantity(magnitude)
        # parse unit
        string_bak = string
        exp, base, prefix = '', '', ''
        symbol, string = string[-1], ' '+string[:-1]
        # parse exponent
        while len(string):
            if not re.match('^[0-9'+Fraction.symbol+'+-]{1}$', symbol):
                break
            exp = symbol+exp
            symbol, string = string[-1], string[:-1]
        # parse unit symbol
        unitkeys = self.unitlist.keys()
        while len(string):
            nbase = len(base)+1
            ukeys = [key[-nbase:] for key in unitkeys]
            if symbol+base not in ukeys:
                break
            base = symbol+base
            symbol, string = string[-1], string[:-1]
        unitid = f"{base:s}"
        if base not in self.unitlist.keys():
            raise Exception('Unknown unit', base)
        magnitude = self.unitlist[base].magnitude
        dimensions = self.unitlist[base].dimensions
        # parse unit prefix
        while len(string):
            prefix = symbol+prefix
            symbol, string = string[-1], string[:-1]
            if symbol==' ':
                break
        if prefix:
            if prefix not in self.prefixes.keys():
                raise Exception(f"Unknown unit prefix:", string_bak)
            magnitude *= self.prefixes[prefix].magnitude
            unitid = f"{prefix:s}{BaseUnits.symbol}{unitid}"
        # apply exponent
        if exp and Fraction.symbol in exp:
            exp = exp.split(Fraction.symbol)
            exprat = Fraction(int(exp[0]), int(exp[1]))
            magnitude = magnitude**(int(exp[0])/int(exp[1]))
            dimensions = [exprat*dim for dim in dimensions]
            baseunits = {unitid: exprat}
        elif exp:
            exp = int(exp)
            magnitude = magnitude**exp
            dimensions = [dim*exp for dim in dimensions]
            baseunits = {unitid: exp}            
        else:
            baseunits = {unitid: 1}
        return Quantity(magnitude, dimensions, baseunits)
    
    def value(self, expression=None):
        if expression:
            return self.to(expression).value()
        elif expr:=self.baseunits.expression():
            return (self/Quantity(1, expr)).magnitude
        else:
            return self.magnitude

    def units(self):
        return self.baseunits.expression()

    def to(self, units: Union[str,list,np.ndarray,Dimensions,dict,BaseUnits]):
        unit1 = self
        unit2 = Quantity(1,units)
        # Check if units can be directly converted
        if not unit1.dimensions==unit2.dimensions:
            # Check if inverted unit can be converted
            if -unit1.dimensions==unit2.dimensions:
                unit1 = Quantity(1)/self
            else:
                raise Exception("Converting units with different dimensions:",
                                unit1.dimensions, unit2.dimensions)
        with TemperatureConverter(unit1.baseunits.value(), unit2.baseunits.value()) as tc:
            if tc.convertable:
                unit2.magnitude = unit1.magnitude/tc.convert(unit1.magnitude, unit2.magnitude)
        unit = unit1/unit2
        return Quantity(unit.magnitude, units)

def implements(np_function):
    def decorator(func):
        HANDLED_FUNCTIONS[np_function] = func
        return func
    return decorator

@implements(np.linspace)
def linspace(a, b, c, **kwargs):
    if isinstance(a,Quantity):
        b = b.to(a.baseunits) if isinstance(b,Quantity) else Quantity(b, a.baseunits)
    else:
        a = a.to(b.baseunits) if isinstance(a,Quantity) else Quantity(a, b.baseunits)
    return Quantity(np.linspace(a.value(), b.value(), c, **kwargs), a.baseunits)

@implements(np.logspace)
def logspace(a, b, c, **kwargs):
    if isinstance(a,Quantity):
        b = b.to(a.baseunits) if isinstance(b,Quantity) else Quantity(b, a.baseunits)
    else:
        a = a.to(b.baseunits) if isinstance(a,Quantity) else Quantity(a, b.baseunits)
    return Quantity(np.logspace(a.value(), b.value(), c, **kwargs), a.baseunits)

@implements(np.absolute)
def absolute(a, **kwargs):
    return Quantity(np.absolute(a.value()), a.baseunits)

@implements(np.abs)
def abs(a, **kwargs):
    return Quantity(np.abs(a.value()), a.baseunits)



from typing import List
import numpy as np
import re
from math import isclose
from typing import Union
from dataclasses import dataclass

from ...structs.ParameterClass import ParameterDict
from ...math.solver import ExpressionSolver, AtomBase, OperatorPar, OperatorMul, OperatorTruediv
from .UnitList import *
from .UnitConverters import TemperatureConverter

class Quantity:
    prefixes: dict            # list of prefixes 
    unitlist: dict            # list of units
    
    magnitude: float          # quantity magnitude
    dimensions: List[int]     # quantity dimensions
    baseunits: dict           # base units

    precision: float = 1e-7

    def __init__(
            self, magnitude:float,
            dimensions = None,
            baseunits = {}
    ):
        # Initialize settings
        self.unitlist = ParameterDict(['magnitude','dimensions','definition','name'], UnitStandard)
        self.prefixes = ParameterDict(['magnitude','dimensions','definition','name'], UnitPrefixes)
        if isinstance(magnitude, (int,float)):
            self.magnitude = float(magnitude)
        elif isinstance(magnitude, list):
            self.magnitude = np.array(magnitude, dtype=float)
        elif isinstance(magnitude, np.ndarray):
            self.magnitude = magnitude.astype(float)
        else:
            raise Exception("Magnitude can be either a number or an list/array of numbers")
        # Set quantity
        if dimensions is None:
            self.dimensions = [0]*len(UnitBase)
            self.baseunits = {}
        elif isinstance(dimensions, str):
            with ExpressionSolver(self._atom_parser, [OperatorPar,OperatorMul,OperatorTruediv]) as es:
                unit = es.solve(dimensions)
            self.magnitude *= unit.magnitude
            self.dimensions = list(unit.dimensions)
            self.baseunits = dict(unit.baseunits)
        elif isinstance(dimensions, (list, np.ndarray)):
            self.dimensions = list(dimensions)
            if baseunits:
                self.baseunits = baseunits
            else:
                self.baseunits = {UnitBase[d]:dim for d,dim in enumerate(self.dimensions) if dim!=0}
        else:
            raise Exception("Insufficient quantity definition", magnitude, dimensions, baseunits)
        # Remove zero base units
        for unit in list(self.baseunits.keys()):
            if self.baseunits[unit]==0:
                del self.baseunits[unit]
        
    def __add__(self, other):
        if not isinstance(other, Quantity):
            other = Quantity(other)
        if not self.dimensions==other.dimensions:
            raise Exception('Addition of two units with different dimensions:', self, other)
        magnitude = self.magnitude + other.magnitude
        dimensions = list(self.dimensions)
        baseunits = dict(self.baseunits)
        return Quantity(magnitude, dimensions, baseunits)

    def __sub__(self, other):
        if not isinstance(other, Quantity):
            other = Quantity(other)
        if not self.dimensions==other.dimensions:
            raise Exception('Substraction of two units with different dimensions:', self, other)
        magnitude = self.magnitude - other.magnitude
        dimensions = list(self.dimensions)
        baseunits = dict(self.baseunits)
        return Quantity(magnitude, dimensions, baseunits)
    
    def __mul__(self, other):
        if not isinstance(other, Quantity):
            other = Quantity(other)
        magnitude = self.magnitude * other.magnitude
        dimensions = [self.dimensions[i]+other.dimensions[i] for i in range(len(UnitBase))]
        baseunits = dict(self.baseunits)
        for unit,exp in other.baseunits.items():
            baseunits[unit] = baseunits[unit]+exp if unit in baseunits else exp
        return Quantity(magnitude, dimensions, baseunits)

    def __truediv__(self, other):
        if not isinstance(other, Quantity):
            other = Quantity(other)
        magnitude = self.magnitude / other.magnitude
        dimensions = [self.dimensions[i]-other.dimensions[i] for i in range(len(UnitBase))]
        baseunits = dict(self.baseunits)
        for unit,exp in other.baseunits.items():
            baseunits[unit] = baseunits[unit]-exp if unit in baseunits else -exp
        return Quantity(magnitude, dimensions, baseunits)

    def __pow__(self, power):
        magnitude = self.magnitude**power
        dimensions = [self.dimensions[i]*power for i in range(len(UnitBase))]
        baseunits = {unit:exp*power for unit,exp in self.baseunits.items()}
        return Quantity(magnitude, dimensions, baseunits)

    def __eq__(self, other):
        if not isclose(self.magnitude, other.magnitude, rel_tol=self.precision):
            return False
        if self.dimensions!=other.dimensions:
            return False
        return True
    
    def __str__(self):
        magnitude = self.value()
        if isinstance(magnitude, np.ndarray):
            with np.printoptions(precision=3, suppress=False, threshold=5):
                magnitude = f"{str(magnitude):s}"
        else:
            magnitude = f"{magnitude:.03e}"
        if self.baseunits:
            return f"Quantity({magnitude:s} {self.units()})"
        else:
            return f"Quantity({magnitude:s})"
            
    def __repr__(self):
        magnitude = self.value()
        if isinstance(magnitude, np.ndarray):
            with np.printoptions(precision=3, suppress=False, threshold=5):
                magnitude = f"{str(magnitude):s}"
        else:
            magnitude = f"{magnitude:.03e}"
        if self.baseunits:
            return f"Quantity({magnitude:s} {self.units()})"
        else:
            return f"Quantity({magnitude:s})"

    def __array__(self):
        return self.magnitude

    def __array_wrap__(self, out_arr, context=None):
        return Quantity(out_arr, self.dimensions, self.baseunits)
    
    def _atom_parser(self, string=None):
        # parse number
        m = re.match(r'^[-]?([0-9.]+)(e([0-9+-]+)|)$', str(string))
        if m:
            magnitude = float(string)
            return Quantity(magnitude)
        # parse unit
        string_bak = string
        exp, base, prefix = '', '', ''
        symbol, string = string[-1], ' '+string[:-1]
        # parse exponent
        while len(string):
            if not re.match('^[0-9+-]{1}$', symbol):
                break
            exp = symbol+exp
            symbol, string = string[-1], string[:-1]
        # parse unit symbol
        unitkeys = self.unitlist.keys()
        while len(string):
            nbase = len(base)+1
            ukeys = [key[-nbase:] for key in unitkeys]
            if symbol+base not in ukeys:
                break
            base = symbol+base
            symbol, string = string[-1], string[:-1]
        unitid = f"{base:s}"
        magnitude = self.unitlist[base].magnitude
        dimensions = self.unitlist[base].dimensions
        # parse unit prefix
        while len(string):
            prefix = symbol+prefix
            symbol, string = string[-1], string[:-1]
            if symbol==' ':
                break
        if prefix:
            if prefix not in self.prefixes.keys():
                raise Exception(f"Unknown unit prefix:", string_bak)
            magnitude *= self.prefixes[prefix].magnitude
            unitid = f"{prefix:s}:{unitid}"
        # apply exponent
        if exp:
            exp = int(exp)
            magnitude = magnitude**exp
            dimensions = [dim*exp for dim in dimensions]
            baseunits = {unitid: exp}
        else:
            baseunits = {unitid: 1}
        return Quantity(magnitude, dimensions, baseunits)
    
    def value(self):
        if self.baseunits:
            unit = self/Quantity(1,self.units(self.baseunits))
        else:
            return self.magnitude
        return unit.magnitude

    def units(self, baseunits=None):
        if baseunits is None:
            baseunits = self.baseunits
        units = []
        for unitid,exponent in baseunits.items():
            symbol = unitid.replace(':','')
            units.append(f"{symbol}" if exponent==1 else f"{symbol}{exponent}")
        return "*".join(units)

    def to(self, units):
        if isinstance(units,str):
            unit1 = self
            unit2 = Quantity(1,units)
            # Check if units can be directly converted
            if unit1.dimensions!=unit2.dimensions:
                # Check if inverted unit can be converted
                if np.all([-unit1.dimensions[d]==unit2.dimensions[d] for d in range(len(UnitBase))]):
                    unit1 = Quantity(1)/self
                else:
                    raise Exception("Converting units with different dimensions:",
                                    unit1.dimensions, unit2.dimensions)
            with TemperatureConverter(unit1.baseunits, unit2.baseunits) as tc:
                if tc.convertable:
                    return Quantity(tc.convert(unit1.magnitude, unit2.magnitude), units)
            unit = unit1/unit2
            return Quantity(unit.magnitude, units)
        else:
            raise Exception("Invalid units format:", units)



import scinumtools.structs
import scinumtools.stats
import scinumtools.data
import scinumtools.phys
from scinumtools.phys.units import Quantity as quant
from scinumtools.phys.units import Unit as unit
from scinumtools.phys.units import Constant as const
